/*
 * inetd/inetd.c --
 *
 *      This module implements the core of an inetd. See the
 *      assignment sheet #6 from the Spring 2011 OS course for the
 *      description.
 *
 * Code by Vladimir Zamdzhiev <v.zamdzhiev@jacobs-university.de>
 * and Juergen Schoenwaelder <j.schoenwaelder@jacobs-university.de>
 */

#define _POSIX_C_SOURCE 201112L
#define _DEFAULT_SOURCE
#define _DARWIN_C_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <errno.h>

#include "inetd.h"

static const char *progname = "inetd";

/*
 * Create a listening TCP endpoint. First get the list of potential
 * network layter addresses and transport layer port numbers. Iterate
 * through the returned address list until an attempt to create a
 * listening TCP endpoint is successful (or no other alternative
 * exists).
 */

static int
tcp_listen(char *port)
{
    struct addrinfo hints, *ai_list, *ai;
    int n, fd = 0, on = 1;
    
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    n = getaddrinfo(NULL, port, &hints, &ai_list);
    if (n) {
	syslog(LOG_ERR, "getaddrinfo() failed for port %s: %s", port,
	       gai_strerror(n));
        return -1;
    }

    for (ai = ai_list; ai; ai = ai->ai_next) {
        fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
        if (fd < 0) {
            continue;
        }

        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
        if (bind(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
            break;
        }
        close(fd);
    }

    freeaddrinfo(ai_list);

    if (ai == NULL) {
        syslog(LOG_ERR, "socket() or bind() failed for port %s", port);
        return -1;
    }

    if (listen(fd, 42) < 0) {
        syslog(LOG_ERR, "listen() failed for port %s: %m", port);
	close(fd);
        return -1;
    }

    return fd;
}

/*
 * Handle a client by allocating a data structure, forking a child
 * process, setting up stdin/stdout/stderr, and executing the
 * requested service program.
 */

static void
handle_client(inetd_service_t *sp)
{
    inetd_client_t *cp;
    struct sockaddr_storage ss;
    socklen_t sslen;
    
    cp = inetd_new_client(sp);

    cp->socket = accept(sp->socket, NULL, NULL);
    if (cp->socket == -1) {
	syslog(LOG_ERR, "accept() failed for port %s - client ignored",
	       sp->port);
	inetd_del_client(cp);
	return;
    }

    cp->pid = fork();
    if (cp->pid == -1) {
	syslog(LOG_ERR, "fork() failed for port %s: %m - client ignored",
	       sp->port);
	inetd_del_client(cp);
	return;
    }

    if (cp->pid == 0) {
	dup2(cp->socket, STDIN_FILENO);
	dup2(cp->socket, STDOUT_FILENO);
	dup2(cp->socket, STDERR_FILENO);
	(void) close(cp->socket);
	execv(sp->name, sp->params);
	syslog(LOG_ERR, "execv() failed: %m - child exiting");
	(void) exit(EXIT_FAILURE);
    } else {
	if (getsockname(cp->socket, (struct sockaddr *) &ss, &sslen) == 0) {
	    getnameinfo((struct sockaddr *) &ss, sslen,
			cp->local_host, sizeof(cp->local_host),
			cp->local_serv, sizeof(cp->local_serv),
			NI_NUMERICHOST);
	}
	if (getpeername(cp->socket, (struct sockaddr *) &ss, &sslen) == 0) {
	    getnameinfo((struct sockaddr *) &ss, sslen,
			cp->peer_host, sizeof(cp->peer_host),
			cp->peer_serv, sizeof(cp->peer_serv),
			NI_NUMERICHOST);
	}
	(void) close(cp->socket);
    }
}

int
main(int argc, char *argv[])
{
    int c, maxfd;
    int daemon = 1;
    char *config_file = "inetd.conf";
    inetd_service_t *sp;
    inetd_client_t *cp;
    fd_set fdset;
    int status;
    pid_t pid;
    
    while ((c = getopt(argc, argv, "fhc:")) >= 0) 
    {
	switch (c) 
	{
	case 'c':
	    config_file = optarg;
	    break;
	case 'f':
	    daemon = 0;
	    break;
	case 'h':
	    printf("Usage: %s [-f] [-c config] [-h]\n", progname);
	    exit(EXIT_SUCCESS);
	}
    }
    
    if (daemon) {
	inetd_daemonize();
    }
    
    openlog(progname, LOG_PID, LOG_DAEMON);
    inetd_signals();
    inetd_read_config(config_file);

    /*
     * Open listening sockets for all configured services.
     */

    for (sp = inetd_services; sp; sp = sp->next) {
	sp->socket = tcp_listen(sp->port);
    }

    while (1) {
	struct timeval tv = { 2, 0 };
	
	FD_ZERO(&fdset);
	for (sp = inetd_services, maxfd = -1; sp; sp = sp->next) {
	    if (sp->socket > -1) FD_SET(sp->socket, &fdset);
	    if (sp->socket > maxfd) maxfd = sp->socket;
	}

	if (maxfd == -1) {
	    break;
	}

	/*
	 * The select() call is time bounded if there are any clients
	 * so that we can reap them once they have exited. This is
	 * simpler than dealing with SIGCHLD signals that can arrive
	 * at any point in time during the execution.
	 */

	if (select(1 + maxfd, &fdset, NULL, NULL,
		   inetd_clients ? &tv : NULL) == -1) {
	    if (errno == EINTR) {
		continue;
	    } 
	    syslog(LOG_ERR, "select() failed: %m - exiting");
	    return EXIT_FAILURE;
        }

	for (sp = inetd_services; sp; sp = sp->next) {
	    if (sp->socket > -1 && FD_ISSET(sp->socket, &fdset)) {
		handle_client(sp);
	    }
	}

	/*
	 * Reap any dead childs here; provide logging information if a
	 * child did not terminate normally, which is something that
	 * should not happen.
	 */
	
	pid = waitpid(-1, &status, WNOHANG);
	for (cp = inetd_clients; cp; cp = cp->next) {
	    if (cp->pid == pid) break;
	}
	if (cp) {
	    if (!WIFEXITED(status) && cp->service) {
		syslog(LOG_ERR, "child for port %s exited abnormally",
		       cp->service->port);
	    }
	    inetd_del_client(cp);
	}
    }

    closelog();
    
    return EXIT_SUCCESS;
}
